import numpy as np
import typing as tp

def rotate(pivot_point: tp.Iterable[float], turning_point: tp.Iterable[float], angle: tp.Iterable[float]) -> np.ndarray: ...
def get_rotate_matrix(dim: int, angle: tp.Iterable[float]) -> np.ndarray: ...
def get_rotate_matrix_2d(angle: float) -> np.ndarray: ...
def get_rotate_matrix_3d(angle: tp.Iterable[float]) -> np.ndarray: ...
